/*
 * sunpos.c
 * kirk johnson
 * july 1993
 *
 * code for calculating the position on the earth's surface for which
 * the sun is directly overhead (adapted from _practical astronomy
 * with your calculator, third edition_, peter duffett-smith,
 * cambridge university press, 1988.)
 *
 * RCS $Id: sunpos.c,v 1.4 1995/09/24 00:51:03 tuna Exp $
 *
 * Copyright (C) 1989, 1990, 1993, 1994, 1995 Kirk Lauritz Johnson
 *
 * Parts of the source code (as marked) are:
 *   Copyright (C) 1989, 1990, 1991 by Jim Frost
 *   Copyright (C) 1992 by Jamie Zawinski <jwz@lucid.com>
 *
 * Permission to use, copy, modify and freely distribute xearth for
 * non-commercial and not-for-profit purposes is hereby granted
 * without fee, provided that both the above copyright notice and this
 * permission notice appear in all copies and in supporting
 * documentation.
 *
 * Unisys Corporation holds worldwide patent rights on the Lempel Zev
 * Welch (LZW) compression technique employed in the CompuServe GIF
 * image file format as well as in other formats. Unisys has made it
 * clear, however, that it does not require licensing or fees to be
 * paid for freely distributed, non-commercial applications (such as
 * xearth) that employ LZW/GIF technology. Those wishing further
 * information about licensing the LZW patent should contact Unisys
 * directly at (lzw_info@unisys.com) or by writing to
 *
 *   Unisys Corporation
 *   Welch Licensing Department
 *   M/S-C1SW19
 *   P.O. Box 500
 *   Blue Bell, PA 19424
 *
 * The author makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "xearth.h"
#include "kljcpyrt.h"

#define TWOPI (2*M_PI)

/*
 * the epoch upon which these astronomical calculations are based is
 * 1990 january 0.0, 631065600 seconds since the beginning of the
 * "unix epoch" (00:00:00 GMT, Jan. 1, 1970)
 *
 * Rats! For Mars, I only have data for a 1975.0 epoch!
 * This epoch is 157680000 seconds since the Unix epoch.
 *
 * given a number of seconds since the start of the unix epoch,
 * EarthDaysSinceEpoch() computes the number of earth days since the start of the
 * astronomical epoch (1990 january 0.0)
 */

#define EpochStart           (157680000)
#define EarthDaysSinceEpoch(secs) (((secs)-EpochStart)*(1.0/(24*3600)))

/*
 * assuming the apparent orbit of the sun about the earth is circular,
 * the rate at which the orbit progresses is given by RadsPerMarsDay --
 * TWOPI radians per orbit divided by 365.242191 days per year:
 * (i.e per tropical year).
 *
 * For Mars this is about 686.98 earth days per martian tropical year. We
 * work in earth days because that's what our computer clocks give us!
 * NB: Remember to alter this when we colonise Mars....
 */

#define RadsPerMarsDay (TWOPI/686.980)

/*
 * details of sun's apparent orbit at epoch 1990.0 (after
 * duffett-smith, table 6, section 46)
 *
 * Epsilon_g    (ecliptic longitude at epoch 1990.0) 279.403303 degrees
 * OmegaBar_g   (ecliptic longitude of perigee)      282.768422 degrees
 * Eccentricity (eccentricity of orbit)                0.016713
 *
 * For Mars, these are 69.62919, 55.59881 and 0.093382 respectively (at
 * epoch 1975.0). But the first two are measured from Earth's vernal equinox
 * as they are intended for earth observation! We want them w.r.t Mars's
 * vernal equinox. So add a constant. We'll guess this for now by setting
 * 13th March 1997 to look like martian northern midsummer day by eye.
 */

#define XX		95.0	/* difference between martian and terrestial
					vernal equinoxes at a guess */
					
#define Epsilon_g    ((XX+69.62919)*(TWOPI/360))
#define OmegaBar_g   ((XX+55.59881)*(TWOPI/360))
#define Eccentricity (0.093382)

/*
 * MeanObliquity gives the mean obliquity of the earth's axis at epoch
 * 1990.0 (computed as 23.440592 degrees according to the method given
 * in duffett-smith, section 27)
 *
 * For Mars, this is about 25.200 degrees.
 */
#define MeanObliquity (25.2000*(TWOPI/360))

static double solve_keplers_equation _P((double));
static double sun_ecliptic_longitude _P((time_t));
static void   ecliptic_to_equatorial _P((double, double, double *, double *));
static double julian_date _P((int, int, int));
static double GST _P((time_t));

/*
 * solve Kepler's equation via Newton's method
 * (after duffett-smith, section 47)
 */
static double solve_keplers_equation(M)
     double M;
{
  double E;
  double delta;

  E = M;
  while (1)
  {
    delta = E - Eccentricity*sin(E) - M;
    if (fabs(delta) <= 1e-10) break;
    E -= delta / (1 - Eccentricity*cos(E));
  }

  return E;
}


/*
 * compute ecliptic longitude of sun (in radians)
 * (after duffett-smith, section 47)
 */
static double sun_ecliptic_longitude(ssue)
     time_t ssue;               /* earth seconds since unix epoch */
{
  double D, N;
  double M_sun, E;
  double v;

  D = EarthDaysSinceEpoch(ssue) * 1.001495;

  N = RadsPerMarsDay * D;
  N = fmod(N, TWOPI);
  if (N < 0) N += TWOPI;

  M_sun = N + Epsilon_g - OmegaBar_g;
  if (M_sun < 0) M_sun += TWOPI;

  E = solve_keplers_equation(M_sun);
  v = 2 * atan(sqrt((1+Eccentricity)/(1-Eccentricity)) * tan(E/2));

  return (v + OmegaBar_g);
}


/*
 * convert from ecliptic to equatorial coordinates
 * (after duffett-smith, section 27)
 */
static void ecliptic_to_equatorial(lambda, beta, alpha, delta)
     double  lambda;            /* ecliptic longitude       */
     double  beta;              /* ecliptic latitude        */
     double *alpha;             /* (return) right ascension */
     double *delta;             /* (return) declination     */
{
  double sin_e, cos_e;

  sin_e = sin(MeanObliquity);
  cos_e = cos(MeanObliquity);

  *alpha = atan2(sin(lambda)*cos_e - tan(beta)*sin_e, cos(lambda));
  *delta = asin(sin(beta)*cos_e + cos(beta)*sin_e*sin(lambda));
}

# ifdef EARTH
/*
 * computing julian dates (assuming gregorian calendar, thus this is
 * only valid for dates of 1582 oct 15 or later)
 * (after duffett-smith, section 4)
 */
static double julian_date(y, m, d)
     int y;                     /* year (e.g. 19xx)          */
     int m;                     /* month (jan=1, feb=2, ...) */
     int d;                     /* day of month              */
{
  int    A, B, C, D;
  double JD;

  /* lazy test to ensure gregorian calendar */
  assert(y >= 1583);

  if ((m == 1) || (m == 2))
  {
    y -= 1;
    m += 12;
  }

  A = y / 100;
  B = 2 - A + (A / 4);
  C = 365.25 * y;
  D = 30.6001 * (m + 1);

  JD = B + C + D + d + 1720994.5;

  return JD;
}
# endif /* EARTH */

/*
 * compute greenwich mean sidereal time (GST) corresponding to a given
 * number of seconds since the unix epoch
 * (after duffett-smith, section 12)
 */
static double GST(ssue)
     time_t ssue;               /* seconds since unix epoch */
{
  double     T0;
# ifdef EARTH
  double     JD;
  double     T;
  double     UT;
  struct tm *tm;

  tm = gmtime(&ssue);

  JD = julian_date(tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday);
  T  = (JD - 2451545) / 36525;

  T0 = ((T + 2.5862e-5) * T + 2400.051336) * T + 6.697374558;

  T0 = fmod(T0, 24.0);
  if (T0 < 0) T0 += 24;

  UT = tm->tm_hour + (tm->tm_min + tm->tm_sec / 60.0) / 60.0;

  T0 += UT * 1.002737909;
# endif /* EARTH */
  /*
   * The equivalent of GST on Mars is PMST (Prime Meridian Sidereal Time)
   * In the absence of a good algorithm, here's a hack!
   * Unix time 867924652L marks 00:00 Prime Meridian Solar Time on Pathfinder
   * Day 0. ( Pathfinder sent its signal at 07:02 Mars *Local* solar time on day
   * 1, which was 21:07GMT Jul 4th 1997. ) We will count sidereal time from this
   * epoch in days of 24hr 37min 22.655sec and add a constant to get things
   * to "look right" for now.
   *
   * The ratio of solar day:sidereal day on mars is 24:39:35.208/24:37:22.655
   * which is 1.001495
   */
   
  T0 = ssue - 867924652L;	/* earth solar seconds since midnight, day 0 */
  T0 /= 1.027491;		/* convert to mars solar seconds */
  T0 *= 1.001495;		/* convert to mars sidereal seconds */
  T0 /= 60*60;			/* convert to mars decimal sidereal hours */
  
  T0 += 23.00;			/* guess difference between solar and sidereal time
  					on pathfinder epoch day (decimal mars hours) */

  T0 = fmod(T0, 24.0);
  if (T0 < 0) T0 += 24;

  return T0;
}


/*
 * given a particular time (expressed in seconds since the unix
 * epoch), compute position on the earth (lat, lon) such that sun is
 * directly overhead.
 *
 * To cope with Mars, the sun-ecliptic longitude is no problem (the constants
 * have been changed), and GST actually returns the Martian Prime Meridian Mean
 * Sidereal Time.
 */
void sun_position(ssue, lat, lon)
     time_t  ssue;              /* seconds since unix epoch */
     double *lat;               /* (return) latitude        */
     double *lon;               /* (return) longitude       */
{
  double lambda;
  double alpha, delta;
  double tmp;

  lambda = sun_ecliptic_longitude(ssue);
  ecliptic_to_equatorial(lambda, 0.0, &alpha, &delta);

  tmp = alpha - (TWOPI/24)*GST(ssue);
  if (tmp < -M_PI)
  {
    do tmp += TWOPI;
    while (tmp < -M_PI);
  }
  else if (tmp > M_PI)
  {
    do tmp -= TWOPI;
    while (tmp < -M_PI);
  }

  *lon = tmp * (360/TWOPI);
  *lat = delta * (360/TWOPI);
}
